#pylint: disable=R0903,F0401,C0103,C0301,R0902,E1101,R0914

"""
This is the client-side part of the password generation protocol.
"""

import requests
import json

# because the client and server are running on the same computer (right now),
# the lib directory is located elsewhere, and we have to hack $PATH to find it
import sys
sys.path.append('../lib')
from pwgen_lib import PublicKey, Commitment, CommittedValue, Password

class GenerateClient(object):
    """Store everything in a wrapper class for convenience"""

    def __init__(self, username, endpoint):
        """The password generation takes place as soon as the class is constructed"""

        self.username = username
        self.endpoint = endpoint
        self.session = requests.Session() # use a session so we have a persisting cookiejar

        resp = self._post(1, {})
        public_key = PublicKey.from_dict(resp['partial_public_key'])
        client_private_key, f_private_key = public_key.modify_private_key()
        data = dict(
            f_client_private_key = f_private_key,
        )

        resp = self._post(2, data)
        commitment_S = Commitment(resp['commitment_S'], public_key)

        committed_C = CommittedValue(None, None, public_key)
        data = dict(
            commitment_C = committed_C.get_commitment(),
        )

        resp = self._post(3, data)
        committed_S = commitment_S.check_and_get_committed_value(resp['reveal_S'], public_key)

        # compute C+S-P and its commitment
        committed_sum = CommittedValue.add(committed_C, committed_S, public_key)
        committed_P = CommittedValue.randomize(committed_sum, public_key)
        inverse_P = CommittedValue.invert(committed_P, public_key)
        committed_zero = CommittedValue.add(committed_sum, inverse_P, public_key)

        data = dict(
            commitment_P = committed_P.get_commitment(),
            commitment_zero = committed_zero.get_commitment(),
            reveal_zero = committed_zero.reveal(),
        )

        resp = self._post(4, data)
        if not resp['password_accepted']:
            raise Exception

        # as described in our protocol paper, the "password" that we need to
        # feed into SRP is actually a slightly modified password based on the
        # private key and committed_P.rnd, so we compute this password (to be
        # sent to the user as a string)
        private_key_inverse = pow(client_private_key + resp['server_private_key'], public_key.q-2, public_key.q)
        for_srp = (committed_P.val + private_key_inverse * committed_P.rnd) % public_key.q
        self.password = Password(num=for_srp)

        # save these as attributes, since other methods need them
        self.committed_P = committed_P
        self.public_key = public_key

    def _post(self, step, payload):
        """Helper method to post data to the password generation service"""

        headers = {'content-type': 'application/json'}
        payload['username'] = self.username
        url = "%s/%s" % (self.endpoint, step)

        r = self.session.post(url, data=json.dumps(payload), headers=headers)
        if r.status_code != requests.codes.ok:
            raise Exception
        return r.json()

    def get_password(self):
        """Return the user's password in plaintext"""
        return self.password.txt

    def get_commitment_P(self):
        """Return the commitment to the user's password"""
        return self.committed_P.get_commitment()

    def get_public_key_dict(self):
        """Return a dict containing the user's public key info"""
        return self.public_key.to_dict()
